#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include "def.h"
#include "dialogbox.h"
#include "input.h"
#include "dfiler.h"

global_option option = {10, 10, 17, "", DEFAULT_FONT_PATH};

SDL_Surface *gScreenSurface;
SDL_Surface *TextSurface;
TTF_Font *font;
TTF_Font *font_small;

SDL_Color color_white = {255, 255, 255, 0};
SDL_Color color_green = {0, 255, 0, 0};
SDL_Color color_red = {255, 0, 0, 0};
SDL_Color color_blue = {0, 0, 255, 0};
SDL_Color color_black = {0, 0, 0, 0};
SDL_Color color_gray = {170, 170, 170, 0};

bool Init(void) {
  FILE *fp;

  if (SDL_Init(SDL_INIT_VIDEO) < 0) return false;

  //  if (SDL_EnableKeyRepeat(300, 50)) return false;  

  gScreenSurface = SDL_SetVideoMode(
				    SCREEN_WIDTH,
				    SCREEN_HEIGHT,
				    SCREEN_BPP,
				    SDL_SWSURFACE// | SDL_FULLSCREEN
				    );

  SDL_ShowCursor(SDL_DISABLE);

  if (TTF_Init()) return false;
  fp = fopen(option.fontfile, "r");
  if (!fp) return false;
  fclose(fp);
  font = TTF_OpenFont(option.fontfile, option.font_size);
  font_small = TTF_OpenFont(option.fontfile, option.font_size-2 < 4 ? 4 : option.font_size-2);

  option.font_height = TTF_FontHeight(font);
  printf("font height = %d\n", option.font_height);
  option.disp_line_max = HEIGHT_MAX/(option.font_height+1) - 2;

  if (!getcwd(option.dfiler_path, sizeof(option.dfiler_path))) {
    printf("dirname of dfiler too long\n");
    return false;
  }

  return true;
}

void End(void) {
  sync();
  printf("sync disc\n");

  if (font) TTF_CloseFont(font);
  if (font_small) TTF_CloseFont(font_small);
  if (gScreenSurface) SDL_FreeSurface(gScreenSurface);

  SDL_Quit();

  return;
}

int CountDir(char *d) {
  DIR *dp;
  struct dirent *entry;
  int i=0;

  dp = opendir(d);
  if (!dp) {
    printf("cannot open dir %s\n", d);
    exit(1);
  }

  while ((entry = readdir(dp)) != NULL) {
    if (strcmp(".",entry->d_name) == 0) {
      continue;
    }

    i++;
  }

  closedir(dp);

  return i;
}

void SeekDir(char *d, char lst[ENTRY_NUM_MAX][ENTRY_NAME_MAX]) {
  DIR *dp;
  struct dirent *entry;
  char buf[ENTRY_NAME_MAX];
  char dl[ENTRY_NUM_MAX/2][ENTRY_NAME_MAX];
  char fl[ENTRY_NUM_MAX/2][ENTRY_NAME_MAX];
  int d_idx=0, f_idx=0, lst_idx=0, i;

  dp = opendir(d);
  if (!dp) {
    printf("error: cannot open dir %s\n", d);
    exit(1);
  }
  while ((entry = readdir(dp)) != NULL) {
    if ((strcmp(".", entry->d_name) == 0)) {
      continue;
    }

    if (entry->d_type == DT_DIR) {
      sprintf(buf, "%s/", entry->d_name);
      strncpy(dl[d_idx], buf, sizeof(dl[d_idx]));
      d_idx++;
      if (d_idx >= ENTRY_NUM_MAX/2) {
	printf("error: too many subdirs\n");
	exit(1);
      }
    } else {
      strncpy(fl[f_idx], entry->d_name, sizeof(fl[d_idx]));
      f_idx++;
      if (f_idx >= ENTRY_NUM_MAX/2) {
	printf("error: too many files\n");
	exit(1);
      }
    }
  }
  closedir(dp);

  qsort(dl, d_idx, sizeof(dl[0]), comp_func);
  qsort(fl, f_idx, sizeof(fl[0]), comp_func);
  for (i=0; i<d_idx; i++) {
    strncpy(lst[lst_idx], dl[i], sizeof(lst[lst_idx]));
    lst_idx++;
  }
  for (i=0; i<f_idx; i++) {
    strncpy(lst[lst_idx], fl[i], sizeof(lst[lst_idx]));
    lst_idx++;
  }

  return;
}

int comp_func(const void *a, const void *b) {
  return(strcmp((char *)a, (char *)b));
}

void BaseName(char *dest, char *src, size_t n) {
  char *head, *tail;

  tail = src;
  while (*tail)
    tail++;
  while (tail > src && tail[-1] == '/')
    tail--;

  head = tail;
  while (head > src && head[-1] != '/')
    head--;

  if (tail-head <= n-1) {
    dest[0] = '\0';
    strncat(dest, head, tail-head);
  } else {
    printf("error: BaseName() buffer full\n");
    exit(1);
  }
  return;
}

int SplitBySpace(char *str, char token[][TOKEN_LENGTH], int token_num) {
  char *p;
  int token_count=0, letter_count=0;

  p = str;
  while (*p != '\0') {
    while ((*p != ' ') && (*p != '\t') && (*p != '\n') && (*p != '\0')) {
      if (letter_count <= TOKEN_LENGTH-2) {
	token[token_count][letter_count] = *p;
	letter_count++;
      }
      p++;
    }
    token[token_count][letter_count] = '\0';
    token_count++; letter_count = 0;
    if (token_count > token_num-1) return token_count;

    while ((*p == ' ') || (*p == '\t') || (*p == '\n')) {
      p++;
    }
  }

  return token_count;
}

bool MainLoop(void) {
  SDL_Event ev;
  SDL_Rect dest, src;
  bool redraw_flag = true;
  int top_line[2] = {0, 0};
  int cursor[2] = {0, 0};
  int current_col = LEFT_COL;
  char entry_list[2][ENTRY_NUM_MAX][ENTRY_NAME_MAX];
  int entry_num[2];
  int draw_num=0;
  char cwd[2][PATHNAME_MAX];
  char tmpstr[ENTRY_NAME_MAX*2+15], tmpstr2[ENTRY_NAME_MAX*2+15];
  int i, j;

  src.x = 0; src.y = 0;
  src.w = WIDTH_MAX; src.h = option.font_height;

  if (!getcwd(cwd[LEFT_COL], sizeof(cwd[LEFT_COL]))) {
    printf("error: pathname too long\n");
    exit(1);
  }
  printf("path: %s\n", cwd[LEFT_COL]);
  strcpy(cwd[RIGHT_COL], cwd[LEFT_COL]);
  for (i=0; i<2; i++) {
    entry_num[i] = CountDir(cwd[i]);
    SeekDir(cwd[i], entry_list[i]);
  }

  while (1) {
    SDL_Delay(10);

    SDL_PollEvent(&ev);
    switch (ev.type){
    case SDL_QUIT:
      return true;
      break;
    }

    GetInput();
    if (button_to_process[PRESS_BTN_START] && button_to_process[PRESS_BTN_SELECT]) {
      return true;
    }
    if (button_to_process[PRESS_BTN_DOWN]) {
      if (cursor[current_col] < option.disp_line_max) {
	if (cursor[current_col] <entry_num[current_col]-1) {
	  cursor[current_col]++;
	}
      } else {
	cursor[current_col] = option.disp_line_max;
	if (top_line[current_col]+option.disp_line_max+1 < entry_num[current_col]) {
	top_line[current_col]++;
	}
      }
      redraw_flag = true;
    }
    if (button_to_process[PRESS_BTN_UP]) {
      if (cursor[current_col] > 0) {
	cursor[current_col]--;
      } else {
	cursor[current_col] = 0;
	if (top_line[current_col] > 0) {
	top_line[current_col]--;
	}
      }
      redraw_flag = true;
    }
    if (button_to_process[PRESS_BTN_RIGHT]) {
      if (current_col == LEFT_COL) {
	current_col = RIGHT_COL;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_LEFT]) {
      if (current_col == RIGHT_COL) {
	current_col = LEFT_COL;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_L]) {
      redraw_flag = true;
    }
    if (button_to_process[PRESS_BTN_R]) {
      redraw_flag = true;
    }
    if (button_to_process[PRESS_BTN_A] && (!button_to_process[PRESS_BTN_SELECT])) {
      /* chdir */
      if (strchr(entry_list[current_col][top_line[current_col]+cursor[current_col]], '/')) {
	chdir(cwd[current_col]);
	chdir(entry_list[current_col][top_line[current_col]+cursor[current_col]]);
	if (!getcwd(cwd[current_col], sizeof(cwd[current_col]))) {
	  printf("error: pathname too long\n");
	  exit(1);
	}
	printf("chdir: %s\n", cwd[current_col]);
	entry_num[current_col] = CountDir(cwd[current_col]);
	SeekDir(cwd[current_col], entry_list[current_col]);
	top_line[current_col] = 0;
	cursor[current_col] = 0;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_A] && button_to_process[PRESS_BTN_SELECT]) {
      /* execute */
      snprintf(tmpstr, sizeof(tmpstr), "execute \"%s\"  Ok?", entry_list[current_col][top_line[current_col]+cursor[current_col]]);
      if (PopupConfirm(150, 100, tmpstr)) {
	snprintf(tmpstr, sizeof(tmpstr), "%s/%s", cwd[current_col], entry_list[current_col][top_line[current_col]+cursor[current_col]]);
	printf("execute: %s\n", tmpstr);
	chdir(cwd[current_col]);
	End();
	execlp("/bin/sh", "/bin/sh", "-c", tmpstr, NULL);

	//try relaunching dfiler when execution fails
	chdir(option.dfiler_path);
	execlp("./dfiler.dge", "./dfiler.dge", NULL);
      }
    }
    if (button_to_process[PRESS_BTN_B] && (!button_to_process[PRESS_BTN_SELECT])) {
      /* parent dir */
      chdir(cwd[current_col]);
      chdir("..");
      if (!getcwd(cwd[current_col], sizeof(cwd[current_col]))) {
	printf("error: pathname too long\n");
	exit(1);
      }
      printf("chdir: %s\n", cwd[current_col]);
      entry_num[current_col] = CountDir(cwd[current_col]);
      SeekDir(cwd[current_col], entry_list[current_col]);
      top_line[current_col] = 0;
      cursor[current_col] = 0;
      redraw_flag = true;
    }
    if (button_to_process[PRESS_BTN_B] && button_to_process[PRESS_BTN_SELECT]) {
      /* delete */
      if (strncmp(entry_list[current_col][top_line[current_col]+cursor[current_col]], "..", 2) != 0) {
	snprintf(tmpstr, sizeof(tmpstr), "delete \"%s\"  Ok?", entry_list[current_col][top_line[current_col]+cursor[current_col]]);
	if (PopupConfirm(150, 100, tmpstr)) {
	  DeleteDirEntry(cwd[current_col], entry_list[current_col][top_line[current_col]+cursor[current_col]]);
	  printf("delete: %s\n", entry_list[current_col][top_line[current_col]+cursor[current_col]]);
	  entry_num[current_col] = CountDir(cwd[current_col]);
	  SeekDir(cwd[current_col], entry_list[current_col]);
	  top_line[current_col] = 0;
	  cursor[current_col] = 0;
	  redraw_flag = true;
	}
      }
    }
    if (button_to_process[PRESS_BTN_Y] && (!button_to_process[PRESS_BTN_SELECT])) {
      /* copy */
      if (current_col == LEFT_COL) i=RIGHT_COL; else i=LEFT_COL;
      if ((strncmp(entry_list[current_col][top_line[current_col]+cursor[current_col]], "..", 2) != 0) && (strcmp(cwd[current_col], cwd[i]) != 0)) {
	printf("copy: %s from %s to %s\n", entry_list[current_col][top_line[current_col]+cursor[current_col]], cwd[current_col], cwd[i]);
	CopyDirEntry(cwd[i], cwd[current_col], entry_list[current_col][top_line[current_col]+cursor[current_col]]);
	entry_num[i] = CountDir(cwd[i]);
	SeekDir(cwd[i], entry_list[i]);
	top_line[i] = 0;
	cursor[i] = 0;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_Y] && button_to_process[PRESS_BTN_SELECT]) {
      /* mkdir */
      LineInput(210, 70, ENTRY_NAME_MAX, 35, "create dir", "", tmpstr);
      chdir(cwd[current_col]);
      if (strlen(tmpstr)) {
	if (access(tmpstr, F_OK) == 0) {
	  snprintf(tmpstr2, sizeof(tmpstr2), "\"%s\" already exists", tmpstr);
	  PopupMessage(120, 80, "Error", tmpstr2);
	} else {
	  if (mkdir(tmpstr, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == -1) {
	    PopupMessage(120, 80, "Error", "mkdir() failed");
	  } else {
	    printf("mkdir: %s in %s\n", tmpstr, cwd[current_col]);
	    entry_num[current_col] = CountDir(cwd[current_col]);
	    SeekDir(cwd[current_col], entry_list[current_col]);
	    top_line[current_col] = 0;
	    cursor[current_col] = 0;
	    redraw_flag = true;
	  }
	}
      }
    }
    if (button_to_process[PRESS_BTN_X] && (!button_to_process[PRESS_BTN_SELECT])) {
      /* move */
      if (current_col == LEFT_COL) i=RIGHT_COL; else i=LEFT_COL;
      if ((strncmp(entry_list[current_col][top_line[current_col]+cursor[current_col]], "..", 2) != 0) && (strcmp(cwd[current_col], cwd[i]) != 0)) {
	printf("move: %s from %s to %s\n", entry_list[current_col][top_line[current_col]+cursor[current_col]], cwd[current_col], cwd[i]);
	MoveDirEntry(cwd[i], cwd[current_col], entry_list[current_col][top_line[current_col]+cursor[current_col]]);
	entry_num[i] = CountDir(cwd[i]);
	SeekDir(cwd[i], entry_list[i]);
	top_line[i] = 0;
	cursor[i] = 0;
	entry_num[current_col] = CountDir(cwd[current_col]);
	SeekDir(cwd[current_col], entry_list[current_col]);
	top_line[current_col] = 0;
	cursor[current_col] = 0;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_X] && button_to_process[PRESS_BTN_SELECT]) {
      /* rename */
      tmpstr[0] = '\0'; strncat(tmpstr, entry_list[current_col][top_line[current_col]+cursor[current_col]], sizeof(tmpstr));
      if (strncmp(tmpstr, "..", 2) != 0) {
	if (tmpstr[strlen(tmpstr)-1] == '/') tmpstr[strlen(tmpstr)-1] = '\0';
	LineInput(210, 70, ENTRY_NAME_MAX, 35, "rename", tmpstr, tmpstr2);
	chdir(cwd[current_col]);
	if ((strlen(tmpstr2) > 0) && strcmp(tmpstr, tmpstr2) != 0) {
	  if (access(tmpstr2, F_OK) == 0) {
	    snprintf(tmpstr, sizeof(tmpstr), "\"%s\" already exists", tmpstr2);
	    PopupMessage(120, 80, "Error", tmpstr);
	  } else {
	    printf("rename: %s to %s\n", tmpstr, tmpstr2);
	    if (rename(tmpstr, tmpstr2) == -1) {
	      PopupMessage(120, 80, "Error", "rename() failed");
	    } else {
	      entry_num[current_col] = CountDir(cwd[current_col]);
	      SeekDir(cwd[current_col], entry_list[current_col]);
	      top_line[current_col] = 0;
	      cursor[current_col] = 0;
	      redraw_flag = true;
	    }
	  }
	}
      }
    }
    if (button_to_process[PRESS_BTN_L]) {
      /* file info */
      snprintf(tmpstr, sizeof(tmpstr), "%s/%s", cwd[current_col], entry_list[current_col][top_line[current_col]+cursor[current_col]]);
      PopupFileInfo(200, 130, tmpstr);
    }
    if (button_to_process[PRESS_BTN_R]) {
      /* disk info */
      PopupDiskInfo(230, 120, cwd[current_col]);
    }

    if (redraw_flag) {
      FillScreen(color_black);
      for (i=0; i<2; i++) {
	if (i == current_col) {
	  TextSurface = TTF_RenderUTF8_Shaded(font, cwd[i], color_blue, color_white);
	} else {
	  TextSurface = TTF_RenderUTF8_Shaded(font, cwd[i], color_blue, color_gray);
	}
	dest.x = 2+(WIDTH_MAX+6)*i;
	dest.y = 3;
	SDL_BlitSurface(TextSurface, NULL, gScreenSurface, &dest);
	SDL_FreeSurface(TextSurface);

	draw_num =  option.disp_line_max > entry_num[i]-1 ? entry_num[i]-1 : option.disp_line_max;
	for (j=0; j<=draw_num; j++) {
	  if ((i == 0 && j == cursor[0]) || (i == 1 && j == cursor[1])) {
	    if (i == current_col) {
	      TextSurface = TTF_RenderUTF8_Shaded(font, entry_list[i][top_line[i]+j], color_black, color_red);
	    } else {
	      TextSurface = TTF_RenderUTF8_Shaded(font, entry_list[i][top_line[i]+j], color_black, color_blue);
	    }
	  } else {
	    if (i == current_col) {
	      TextSurface = TTF_RenderUTF8_Shaded(font, entry_list[i][top_line[i]+j], color_black, color_white);
	    } else {
	      TextSurface = TTF_RenderUTF8_Shaded(font, entry_list[i][top_line[i]+j], color_black, color_gray);
	    }
	  }
	  dest.x = 2+(WIDTH_MAX+6)*i;
	  dest.y = 3+(j+1)*(option.font_height+1);
	  SDL_BlitSurface(TextSurface, &src, gScreenSurface, &dest);
	  SDL_FreeSurface(TextSurface);
	}
      }

      redraw_flag = false;
    }

    if (!button_to_process[PRESS_BTN_SELECT]) {
      TextSurface = TTF_RenderUTF8_Shaded(font_small, "A: chdir  B: up     X: move   Y: copy   L: info   R: disk   START:        ", color_black, color_white);
    } else {
      TextSurface = TTF_RenderUTF8_Shaded(font_small, "A: exec   B: delete X: rename Y: mkdir  L:        R:        START: quit   ", color_black, color_white);
    }
    dest.x = 2;
    dest.y = SCREEN_HEIGHT-10;
    SDL_BlitSurface(TextSurface, NULL, gScreenSurface, &dest);
    SDL_FreeSurface(TextSurface);

    SDL_Flip(gScreenSurface);
  }

  return false;
}

void FillScreen(SDL_Color color) {
  SDL_Rect dest;

  dest.x = 0; dest.y = 0;
  dest.w = SCREEN_WIDTH;
  dest.h = SCREEN_HEIGHT;
  SDL_FillRect(gScreenSurface, &dest, SDL_MapRGB(gScreenSurface->format, color.r, color.g, color.b));
}

bool PopupFileInfo(int w, int h, char *fn) {
  SDL_Rect dest;
  SDL_Surface *tmpSurface;
  int width;
  char tmpstr[ENTRY_NAME_MAX];
  struct stat st;
  struct tm mtime;

  if (w > SCREEN_WIDTH || h > SCREEN_HEIGHT) {
    printf("error: too large popup\n");
    return false;
  }

  tmpSurface = BackupScreenSurface();
  DrawDialogBox(-1, -1, w, h, "File Info");

  stat(fn, &st);
  gmtime_r(&(st.st_mtime), &mtime);
  snprintf(tmpstr, sizeof(tmpstr), "%s\n %d bytes\n %d-%02d-%02d %02d:%02d", fn, (int)st.st_size, 1900+mtime.tm_year, 1+mtime.tm_mon, mtime.tm_mday, mtime.tm_hour, mtime.tm_min);
  dest.x = (SCREEN_WIDTH-w)/2 + 10;
  dest.y = (SCREEN_HEIGHT-h)/2 + 10 + option.font_height;
  dest.w = w-20;
  dest.h = h-20-option.font_height;
  DrawText(dest, tmpstr);

  TTF_SizeUTF8(font, "  Ok  ", &width, NULL);
  dest.x = (SCREEN_WIDTH-w)/2 + 5 + (w-10-width)/2;
  dest.y = SCREEN_HEIGHT - (SCREEN_HEIGHT-h)/2 - 10 - option.font_height;
  DrawButton(dest, color_white, color_red, "  Ok  ");

  SDL_Flip(gScreenSurface);

  WaitKey();

  RevertScreenSurface(tmpSurface);

  return true;
}

bool PopupDiskInfo(int w, int h, char *d) {
  SDL_Rect dest;
  SDL_Surface *tmpSurface;
  FILE *pipe;
  char buf[ENTRY_NAME_MAX+6], buf2[ENTRY_NAME_MAX+6];
  const int tnum = 6;
  char token[tnum][TOKEN_LENGTH];
  int width;

  if (w > SCREEN_WIDTH || h > SCREEN_HEIGHT) {
    printf("error: too large popup\n");
    return false;
  }

  tmpSurface = BackupScreenSurface();

  DrawDialogBox(-1, -1, w, h, "Disk Info");
  TTF_SizeUTF8(font, "Reading...", &width, NULL);
  dest.x = (SCREEN_WIDTH-w)/2 + (w-width)/2;
  dest.y = (SCREEN_HEIGHT-h)/2 + h/2;
  dest.w = width+10;
  dest.h = option.font_height+10;
  DrawText(dest, "Reading...");
  SDL_Flip(gScreenSurface);

  DrawDialogBox(-1, -1, w, h, "Disk Info");
  pipe = popen("df -h", "r");
  while (fgets(buf, sizeof(buf), pipe) != NULL) {
    if (strstr(buf, DISK_MOUNTPOINT) != NULL) {
      SplitBySpace(buf, token, tnum);
    }
  }
  pclose(pipe);
  snprintf(buf, sizeof(buf), "Size: %s  Used: %s  Avail: %s\n  %s used", token[1], token[2], token[3], token[4]);
  dest.x = (SCREEN_WIDTH-w)/2 + 10;
  dest.y = (SCREEN_HEIGHT-h)/2 + 10 + option.font_height;
  dest.w = w-20;
  dest.h = (h-20-option.font_height)/2;
  DrawText(dest, buf);

  snprintf(buf, sizeof(buf), "du -h %s", d);
  pipe = popen(buf, "r");
  while (fgets(buf, sizeof(buf), pipe) != NULL) {
    strcpy(buf2, buf);
  }
  pclose(pipe);
  SplitBySpace(buf2, token, tnum);
  snprintf(buf, sizeof(buf), "%s\n  total %s", token[1], token[0]);
  dest.y += dest.h;
  DrawText(dest, buf);

  TTF_SizeUTF8(font, "  Ok  ", &width, NULL);
  dest.x = (SCREEN_WIDTH-w)/2 + (w-width)/2;
  dest.y = SCREEN_HEIGHT - (SCREEN_HEIGHT-h)/2 - 10 - option.font_height;
  DrawButton(dest, color_white, color_red, "  Ok  ");

  SDL_Flip(gScreenSurface);

  WaitKey();

  RevertScreenSurface(tmpSurface);

  return true;
}

bool CopyFile(char *dest, char *src) {
  FILE *fp_in, *fp_out;
  char buf[FILE_COPY_BUFFER_SIZE];
  struct stat st;
  int progress_block, processed_size=0, processed_block=0;
  SDL_Rect r;
  char tmpstr[PATHNAME_MAX];
  int numbytes = 0;

  fp_in = fopen(src, "r");
  if (!fp_in) {
    snprintf(tmpstr, sizeof(tmpstr), "can't open %s", src);
    PopupMessage(150, 100, "Error", tmpstr);
    return false;
  }
  fp_out = fopen(dest, "w");
  if (!fp_out) {
    snprintf(tmpstr, sizeof(tmpstr), "can't create %s", dest);
    PopupMessage(150, 100, "Error", tmpstr);
    return false;
  }

  stat(src, &st);
  progress_block = (int)st.st_size/20;

  DrawDialogBox(-1, -1, 200, 80, "Copying");
  BaseName(tmpstr, src, sizeof(tmpstr));
  r.x = 70; r.y = 95+option.font_height;
  r.w = 180; r.h = 60-option.font_height;
  DrawText(r, tmpstr);
  /* progress bar */
  r.x = 70; r.y = 100+2*option.font_height;
  r.w = 180; r.h = 10;
  SDL_FillRect(gScreenSurface, &r, SDL_MapRGB(gScreenSurface->format, 50, 50, 50));

  SDL_Flip(gScreenSurface);

  while (feof(fp_in) == 0) {
    numbytes = fread(buf, 1, sizeof(buf), fp_in);
    if (ferror(fp_in) != 0) {
      snprintf(tmpstr, sizeof(tmpstr), "read %s failed", src);
      PopupMessage(150, 100, "Error", tmpstr);
      return false;
    }
    fwrite(buf, 1, numbytes, fp_out);
    if (ferror(fp_out) != 0) {
      snprintf(tmpstr, sizeof(tmpstr), "write %s failed", dest);
      PopupMessage(150, 100, "Error", tmpstr);
      return false;
    }
    processed_size += numbytes;
    while (processed_size > progress_block) {
      r.x = 70+9*processed_block; r.y = 100+2*option.font_height;
      r.w = 9; r.h = 10;
      SDL_FillRect(gScreenSurface, &r, SDL_MapRGB(gScreenSurface->format, 255, 255, 255));
      
      SDL_Flip(gScreenSurface);
      processed_block++;
      processed_size -= progress_block;
    }
  }
  fclose(fp_in); fclose(fp_out);

  return true;
}

bool CopyDirEntry(char *destdir, char *srcdir, char *entry) {
  char lst[ENTRY_NUM_MAX][ENTRY_NAME_MAX];
  char srcpath[PATHNAME_MAX];
  char destpath[PATHNAME_MAX];
  int dstcount = 0, i;
  char tmpstr[PATHNAME_MAX];
  struct stat st;

  if (strlen(destdir)+strlen(entry)+2 > PATHNAME_MAX) {
    PopupMessage(120, 80, "Error", "path too long");
    return false;
  }
  snprintf(destpath, sizeof(destpath), "%s/%s", destdir, entry);
  if (stat(destpath, &st) == 0) {
    snprintf(tmpstr, sizeof(tmpstr), "%s already exists", destpath);
    PopupMessage(150, 100, "Error", tmpstr);
    return false;
  }

  if (strlen(srcdir)+strlen(entry)+2 > PATHNAME_MAX) {
    PopupMessage(120, 80, "Error", "path too long");
    return false;
  }
  snprintf(srcpath, sizeof(srcpath), "%s/%s", srcdir, entry);

  stat(srcpath, &st);
  if (S_ISDIR(st.st_mode)) {
    if (mkdir(destpath, S_IRUSR | S_IRGRP | S_IROTH | S_IXUSR | S_IXGRP | S_IXOTH | S_IWUSR | S_IWGRP ) == -1) {
      snprintf(tmpstr, sizeof(tmpstr), "can't create dir %s", destpath);
      PopupMessage(150, 100, "Error", tmpstr);
      return false;
    }
    dstcount = CountDir(srcpath);
    SeekDir(srcpath, lst);
    for (i=0; i<dstcount; i++) {
      if (strcmp("../", lst[i]) == 0) {
	continue;
      }
      if (!CopyDirEntry(destpath, srcpath, lst[i])) return false;
    }
  } else if (S_ISREG(st.st_mode)) {
    if (!CopyFile(destpath, srcpath)) return false;
  } else {
    PopupMessage(120, 80, "Error", "unknown file type");
    return false;
  }

  return true;
}

bool MoveDirEntry(char *destdir, char *srcdir, char *entry) {
  char srcpath[PATHNAME_MAX];
  char destpath[PATHNAME_MAX];
  char tmpstr[PATHNAME_MAX];
  struct stat st;

  if (strlen(destdir)+strlen(entry)+2 > PATHNAME_MAX) {
    PopupMessage(120, 80, "Error", "path too long");
    return false;
  }
  snprintf(destpath, sizeof(destpath), "%s/%s", destdir, entry);
  if (stat(destpath, &st) == 0) {
    snprintf(tmpstr, sizeof(tmpstr), "%s already exists", destpath);
    PopupMessage(150, 100, "Error", tmpstr);
    return false;
  }

  if (strlen(srcdir)+strlen(entry)+2 > PATHNAME_MAX) {
    PopupMessage(120, 80, "Error", "path too long");
    return false;
  }
  snprintf(srcpath, sizeof(srcpath), "%s/%s", srcdir, entry);

  if (rename(srcpath, destpath) == -1) {
    PopupMessage(120, 80, "Error", "rename() failed");
    return false;
  }

  return true;
}

bool DeleteDirEntry(char *d, char *entry) {
  char lst[ENTRY_NUM_MAX][ENTRY_NAME_MAX];
  char path[PATHNAME_MAX];
  int dstcount, i;
  struct stat st;

  if (strlen(d)+strlen(entry)+2 > PATHNAME_MAX) {
    PopupMessage(120, 80, "Error", "path too long");
    return false;
  }
  snprintf(path, sizeof(path), "%s/%s", d, entry);

  stat(path, &st);
  if (S_ISDIR(st.st_mode)) {
    dstcount = CountDir(path);
    SeekDir(path, lst);
    for (i=0; i<dstcount; i++) {
      if (strcmp("../", lst[i]) == 0) {
	continue;
      }
      if (!DeleteDirEntry(path, lst[i])) return false;
    }
    if (rmdir(path) == -1) {
      PopupMessage(120, 80, "Error", "rmdir() failed");
      return false;
    }
  } else if (S_ISREG(st.st_mode)) {
    if (unlink(path) == -1) {
      PopupMessage(120, 80, "Error", "unlink() failed");
      return false;
    }
  } else {
    PopupMessage(120, 80, "Error", "unknown file type");
    return false;
  }

  return true;
}

void Usage(void) {
  fprintf(stderr,"Usage: dfiler [option]\n");
  fprintf(stderr," -s #         : font size\n");
  fprintf(stderr," -f fontfile  : font file\n");
}

int main(int argc, char *argv[]) {
  int ch;
  extern char *optarg;
  
  while ((ch = getopt(argc, argv, "s:f:")) != -1){
    switch (ch){
    case 's':
      if (atoi(optarg) <= 14 && atoi(optarg) >= 8) {
	printf("font size: %d\n", atoi(optarg));
	option.font_size = atoi(optarg);
      }
      break;
    case 'f':
      printf("font file: %s\n",optarg);
      option.fontfile[0] = '\0';
      strncat(option.fontfile, optarg, sizeof(option.fontfile)-1);
      break;
    default:
      Usage();
      return 1;
    }
  }

  if (!Init()) {
    printf("error: init failed\n");
    return 1;
  }

  MainLoop();
  End();

  return 0;
}
