#include <stdio.h>
#include "def.h"
#include "dialogbox.h"
#include "input.h"
#include "dingux_danzeff.h"

SDL_Surface *BackupScreenSurface(void) {
  return SDL_ConvertSurface(gScreenSurface, gScreenSurface->format, SDL_SWSURFACE);
}

void RevertScreenSurface(SDL_Surface *s) {
  SDL_Rect dest;

  dest.x = 0; dest.y = 0;
  dest.w = SCREEN_WIDTH;
  dest.h = SCREEN_HEIGHT;
  SDL_BlitSurface(s, NULL, gScreenSurface, &dest);
  SDL_Flip(gScreenSurface);
  SDL_FreeSurface(s);

  return;
}

int FoldString(int w, int h, char *str, char buf[SCREEN_HEIGHT/4][SCREEN_WIDTH/4]) {
  char *src_p;
  int bytes, line_idx=0, str_idx=0, width, i;

  src_p = str;

  while (*src_p != '\0') {
    if (*src_p == '\n') {
      if ((line_idx+1)*option.font_height >= h) break;
      src_p++;
      line_idx++;
      str_idx = 0;      
    }

    if ((unsigned char)*src_p <= 0x7f) {
      bytes = 1;
    } else if ((unsigned char)*src_p <= 0xbf) {
      printf("error: splitted letter\n");
      return -1;
    } else if ((unsigned char)*src_p <= 0xdf) {
      bytes = 2;
    } else if ((unsigned char)*src_p <= 0xef) {
      bytes = 3;
    } else if ((unsigned char)*src_p <= 0xf7) {
      bytes = 4;
    } else {
      printf("error: unknown letter\n");
      return -1;
    }

    for (i=0; i<bytes; i++) {
      buf[line_idx][str_idx+i] = *src_p;
      src_p++;
    }
    str_idx += bytes;
    buf[line_idx][str_idx] = '\0';

    TTF_SizeUTF8(font, buf[line_idx], &width, NULL);
    if (width > w) {
      buf[line_idx][str_idx-bytes] = '\0';
      src_p -= bytes;
      if ((line_idx+1)*option.font_height >= h) break;
      line_idx++;
      str_idx = 0;
    }
  }

  return line_idx;
}

void DrawDialogBox(int x, int y, int w, int h, char *title) {
  SDL_Rect dest;
  SDL_Surface *dlgSurface;
  int width;

  dlgSurface = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, SCREEN_BPP, 0, 0, 0, 0);

  SDL_FillRect(dlgSurface, NULL, SDL_MapRGB(dlgSurface->format, 100, 100, 255));
  dest.x = 5; dest.y = 5;
  dest.w = w-10; dest.h = h-10;
  SDL_FillRect(dlgSurface, &dest, SDL_MapRGB(dlgSurface->format, 230, 230, 230));

  TTF_SizeUTF8(font, title, &width, NULL);
  dest.x += ((dest.w-width)/2); dest.y += 3;
  TextSurface = TTF_RenderUTF8_Shaded(font, title, color_black, color_gray);
  SDL_BlitSurface(TextSurface, NULL, dlgSurface, &dest);
  SDL_FreeSurface(TextSurface);

  if (x == -1) {
    dest.x = (SCREEN_WIDTH-w)/2;
  } else {
    dest.x = x;
  }
  if (y == -1) {
    dest.y = (SCREEN_HEIGHT-h)/2;
  } else {
    dest.y = y;
  }
  SDL_BlitSurface(dlgSurface, NULL, gScreenSurface, &dest);
  SDL_FreeSurface(dlgSurface);

  return;
}

void DrawButton(SDL_Rect r, SDL_Color fg, SDL_Color bg, char *msg) {
  TextSurface = TTF_RenderUTF8_Shaded(font, msg, fg, bg);
  SDL_BlitSurface(TextSurface, NULL, gScreenSurface, &r);
  SDL_FreeSurface(TextSurface);
  
  return;
}

void DrawText(SDL_Rect r, char *buf) {
  int i, line_num;
  char strbuf[SCREEN_HEIGHT/4][SCREEN_WIDTH/4];

  line_num = FoldString(r.w, r.h, buf, strbuf);
  for (i=0; i<=line_num; i++) {
    TextSurface = TTF_RenderUTF8_Blended(font, strbuf[i], color_black);
    SDL_BlitSurface(TextSurface, NULL, gScreenSurface, &r);
    SDL_FreeSurface(TextSurface);
    r.y += option.font_height;
  }

  return;
}

bool PopupMessage(int w, int h, char *title, char *msg) {
  SDL_Rect dest;
  SDL_Surface *tmpSurface;
  int width;

  if (w > SCREEN_WIDTH || h > SCREEN_HEIGHT) {
    printf("error: too large popup\n");
    return false;
  }

  tmpSurface = BackupScreenSurface();

  DrawDialogBox(-1, -1, w, h, title);
  dest.x = (SCREEN_WIDTH-w)/2 + 10;
  dest.y = (SCREEN_HEIGHT-h)/2 + 10 + option.font_height;
  dest.w = w-20;
  dest.h = h-20-option.font_height*2;
  DrawText(dest, msg);
  TTF_SizeUTF8(font, "  Ok  ", &width, NULL);
  dest.x = (SCREEN_WIDTH-w)/2 + 5 + (w-10-width)/2;
  dest.y = SCREEN_HEIGHT - (SCREEN_HEIGHT-h)/2 - 10 - option.font_height;
  DrawButton(dest, color_white, color_red, "  Ok  ");

  SDL_Flip(gScreenSurface);

  WaitKey();

  RevertScreenSurface(tmpSurface);

  return true;
}

bool PopupConfirm(int w, int h, char *msg) {
  SDL_Rect dest, button_dest[2];
  SDL_Surface *tmpSurface;
  int width;
  bool done=false, answer=false;

  if (w > SCREEN_WIDTH || h > SCREEN_HEIGHT) {
    printf("error: too large popup\n");
    return false;
  }

  tmpSurface = BackupScreenSurface();

  DrawDialogBox(-1, -1, w, h, "Confirm");
  dest.x = (SCREEN_WIDTH-w)/2 + 10;
  dest.y = (SCREEN_HEIGHT-h)/2 + 10 + option.font_height;
  dest.w = w-20;
  dest.h = h-20-option.font_height*2;
  DrawText(dest, msg);

  button_dest[0].x = (SCREEN_WIDTH-w)/2 + 15;
  button_dest[0].y = SCREEN_HEIGHT - (SCREEN_HEIGHT-h)/2 - 10 - option.font_height;
  TTF_SizeUTF8(font, "  Ok  ", &width, NULL);
  button_dest[1].x = SCREEN_WIDTH - (SCREEN_WIDTH-w)/2 - 15 - width;
  button_dest[1].y = SCREEN_HEIGHT - (SCREEN_HEIGHT-h)/2 - 10 - option.font_height;
  DrawButton(button_dest[0], color_white, color_red, "Cancel");
  DrawButton(button_dest[1], color_black, color_gray, "  Ok  ");

  SDL_Flip(gScreenSurface);

  while (!done) {
    SDL_Delay(10);
    GetInput();
    if (button_to_process[PRESS_BTN_LEFT]) {
      if (answer == true) {
	answer = false;
	DrawButton(button_dest[0], color_white, color_red, "Cancel");
	DrawButton(button_dest[1], color_black, color_gray, "  Ok  ");
	SDL_Flip(gScreenSurface);
      }
    }
    if (button_to_process[PRESS_BTN_RIGHT]) {
      if (answer == false) {
	answer = true;
	DrawButton(button_dest[0], color_black, color_gray, "Cancel");
	DrawButton(button_dest[1], color_white, color_red, "  Ok  ");
	SDL_Flip(gScreenSurface);
      }
    }
    if (button_to_process[PRESS_BTN_A]) {
      done = true;
    }
    if (button_to_process[PRESS_BTN_B]) {
      answer = false;
      done  = true;
    }
    SDL_PumpEvents();
  }

  RevertScreenSurface(tmpSurface);

  return answer;
}

bool LineInput(int w, int h, int maxlen, int winsize, char *title, char *str, char *dst) {
  SDL_Rect dest;
  SDL_Surface *tmpSurface;
  int width, col;
  unsigned int danzeff_key;
  char buf[maxlen], tmpbuf[winsize], dir[256];
  bool done=false;

  if (w > SCREEN_WIDTH || h > SCREEN_HEIGHT) {
    printf("error: too large popup\n");
    return false;
  }

  tmpSurface = BackupScreenSurface();

  DrawDialogBox(5, 5, w, h, title);
  strcpy(dir, option.dfiler_path);
  strcat(dir, "/graphics/");
  danzeff_load(dir);
  danzeff_set_screen(gScreenSurface);
  danzeff_moveTo(-10, -10);
  
  buf[0] = '\0'; strncat(buf, str, maxlen-1); buf[maxlen] = '\0';
  col = strlen(buf);

  WaitKeyRelease();
  while (!done) {
    SDL_Delay(10);
    danzeff_key = danzeff_readInput();

    if (danzeff_key >= ' ') {
      if (col >= maxlen-1) {
	col = maxlen-1;
	buf[col] = danzeff_key;
      } else {
	buf[col] = danzeff_key;
	col++;
	buf[col] = '\0';
      }
    } else if (danzeff_key == DANZEFF_DEL) {
      if (col < maxlen-1) {
	buf[col-1] = '\0';
	col--;
      } else {
	if (buf[col] == '\0') {
	  buf[col-1] = '\0';
	  col--;
	} else {
	  buf[col] = '\0';
	}
      }
    } else if (danzeff_key == DANZEFF_CLEAR) {
      buf[0] = '\0';
      col = 0;
    } else if ((danzeff_key == DANZEFF_START) || (danzeff_key == DANZEFF_ENTER)) {
      done = true;
    }

    //TextArea
    dest.x = 20; dest.y = 15+option.font_height+5;
    dest.w = w-30; dest.h = option.font_height;
    SDL_FillRect(gScreenSurface, &dest, SDL_MapRGB(gScreenSurface->format, 0, 0, 0));

    tmpbuf[0] = '\0';
    if (col > winsize) {
      strncat(tmpbuf, &(buf[col-winsize]), winsize);
    } else {
      strncat(tmpbuf, buf, winsize);
    }
    TextSurface = TTF_RenderUTF8_Shaded(font, tmpbuf, color_white, color_black);
    TTF_SizeUTF8(font, tmpbuf, &width, NULL);
    dest.x = 20;
    dest.y = 15+option.font_height+5;
    SDL_BlitSurface(TextSurface, NULL, gScreenSurface, &dest);
    SDL_FreeSurface(TextSurface);

    //cursor
    tmpbuf[0] = (buf[col] == '\0') ? ' ' : buf[col];
    tmpbuf[1] = '\0';
    TextSurface = TTF_RenderUTF8_Shaded(font, tmpbuf, color_white, color_red);
    dest.x = 20+width;
    SDL_BlitSurface(TextSurface, NULL, gScreenSurface, &dest);
    SDL_FreeSurface(TextSurface);

    danzeff_render(true);
    SDL_Flip(gScreenSurface);

    SDL_PumpEvents();
  }
  strcpy(dst, buf);

  danzeff_free();
  RevertScreenSurface(tmpSurface);

  WaitKeyRelease();

  return true;
}
