#include <stdio.h>
#include "def.h"
#include "input.h"

bool button_to_process[PRESS_BTN_MAX] = {false, false, false, false, false, false, false, false, false, false, false, false};
int repeat_mode[PRESS_BTN_MAX] = {REP_REPEAT, REP_REPEAT, REP_REPEAT, REP_REPEAT, REP_NONE, REP_NONE, REP_NONE, REP_NONE, REP_NONE, REP_NONE, REP_RAW, REP_RAW};

void GetInput(void) {
  static int delay[PRESS_BTN_MAX] = {REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY};
  static int current_btn[PRESS_BTN_MAX] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  static int prev_btn[PRESS_BTN_MAX];
  int i;
  Uint8 *keys;

  for (i=0; i<PRESS_BTN_MAX; i++) {
    prev_btn[i] = current_btn[i];
    current_btn[i] = 0;
  }
  keys = SDL_GetKeyState(NULL);
  current_btn[PRESS_BTN_UP]     = keys[SDLK_UP];
  current_btn[PRESS_BTN_DOWN]   = keys[SDLK_DOWN];
  current_btn[PRESS_BTN_LEFT]   = keys[SDLK_LEFT];
  current_btn[PRESS_BTN_RIGHT]  = keys[SDLK_RIGHT];
  current_btn[PRESS_BTN_A]      = keys[SDLK_LCTRL];
  current_btn[PRESS_BTN_B]      = keys[SDLK_LALT];
  current_btn[PRESS_BTN_X]      = keys[SDLK_SPACE];
  current_btn[PRESS_BTN_Y]      = keys[SDLK_LSHIFT];
  current_btn[PRESS_BTN_L]      = keys[SDLK_TAB];
  current_btn[PRESS_BTN_R]      = keys[SDLK_BACKSPACE];
  current_btn[PRESS_BTN_START]  = keys[SDLK_RETURN];
  current_btn[PRESS_BTN_SELECT] = keys[SDLK_ESCAPE];

  for (i=0; i<PRESS_BTN_MAX; i++) {
    switch (repeat_mode[i]) {
    case REP_NONE:
      if (current_btn[i]) {
	if (prev_btn[i] == 0) {
	  button_to_process[i] = true;
	} else {
	  button_to_process[i] = false;
	}
      } else {
	button_to_process[i] = false;
      }
      break;
    case REP_REPEAT:
      if (current_btn[i]) {
	if (prev_btn[i]) {
	  delay[i]--;
	  if (delay[i] <= 0) {
	    button_to_process[i] = true;
	    delay[i] = REPEAT_INTERVAL;
	  } else {
	    button_to_process[i] = false;
	  }
	} else {
	  button_to_process[i] = true;
	  delay[i] = REPEAT_DELAY;
	}
      } else {
	button_to_process[i] = false;
	delay[i] = REPEAT_DELAY;
      }
      break;
    case REP_RAW:
      if (current_btn[i]) {
	button_to_process[i] = true;
      } else {
	button_to_process[i] = false;
      }
      break;
    }
  }

  return;
}

void WaitKeyRelease(void) {
  bool done=false;
  SDL_Event ev;

  while (!done) {
    GetInput();
    done = (button_to_process[PRESS_BTN_LEFT] || button_to_process[PRESS_BTN_RIGHT] || button_to_process[PRESS_BTN_UP] || button_to_process[PRESS_BTN_DOWN] || button_to_process[PRESS_BTN_A] || button_to_process[PRESS_BTN_B] || button_to_process[PRESS_BTN_X] || button_to_process[PRESS_BTN_Y] || button_to_process[PRESS_BTN_L] || button_to_process[PRESS_BTN_R] || button_to_process[PRESS_BTN_START] || button_to_process[PRESS_BTN_SELECT]) ? false : true;
    SDL_PumpEvents();
  }

  // discard SDL events
  while(SDL_PollEvent(&ev)){
    continue;
  }  

  return;
}

void WaitKey(void) {
  bool done=false;

  WaitKeyRelease();

  while (!done) {
    SDL_Delay(10);
    GetInput();
    if (button_to_process[PRESS_BTN_A] || button_to_process[PRESS_BTN_B]) {
      done = true;
    }
    SDL_PumpEvents();
  }

  return;
}

